# -*- coding: utf-8 -*-
r"""
Constants for the shellcommands library
"""
# SpinD path
SPIND_WINDOWS = r'C:\Users\Hendrik\source\repos\SpinD_Kiel\spin.exe'
SPIND_LINUX = r'/home/schrautzer/code/SpinD_Kiel/spin'